# -*- coding: utf-8 -*-


import socket
import struct
import traceback
import random
import time


class Transport(object):
    """
    """

    sock = None
    encoding = 'utf-8'

    def send(self, data):
        """
        send data with length in 4 bytes header
        """
        data = data.encode(self.encoding)
        header = struct.pack('<L', len(data))
        self.sock.sendall(header + data)

    def recv(self):
        """
        recv data
        """
        header = self.recv_bytes(4)
        if header == '':
            return header
        try:
            length = struct.unpack('<L', header)[0]
        except:
            return None
        data = self.recv_bytes(length).decode(self.encoding)
        return data

    def recv_bytes(self, n):
        """
        recv n bytes
        """
        data = b''
        while len(data) < n:
            chunk = self.sock.recv(n - len(data))
            if not chunk:
                break
            data += chunk
        return data

    def close(self):
        try:
            self.sock and self.sock.close()
        except:
            traceback.print_exc()


class SockClient(Transport):
    """
    """

    sock = None
    CONNECT_RETRIES = 3
    CONNECT_RETRY_INTERVAL = 1

    def __init__(self, host, port, encoding='utf-8'):
        self.host = host
        self.port = port
        self.encoding = encoding

    def connect(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        e = None
        for i in range(self.CONNECT_RETRIES):
            try:
                sock.connect((self.host, self.port))
                break
            except socket.error as e:
                pass
            time.sleep(self.CONNECT_RETRY_INTERVAL)
        else:
            if e:
                raise e
        self.sock = sock


class SockServer(Transport):
    """
    """

    RANDOM_PORT_RANGE = (65000, 65500)
    BIND_RETRIES = 10
    ACCEPT_CLIENT_NUM = 1

    host = "127.0.0.1"
    port = None
    server = None
    client = None

    def __init__(self, listen_timeout=30, transport_timeout=0, encoding='utf-8'):
        """
        """
        self.listen_timeout = listen_timeout
        self.transport_timeout = transport_timeout
        self.encoding = encoding
        self.server = None
        self.sock = None

    def listen(self):
        """
        bind and listen
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        for i in range(self.BIND_RETRIES):
            port = self.get_port()
            try:
                sock.bind((self.host, port))
                self.port = port
                sock.settimeout(self.listen_timeout)
                sock.listen(self.ACCEPT_CLIENT_NUM)
                break
            except socket.error as e:
                # not a address already in use exception
                if e.errno != 98:
                    raise e
        else:
            raise NoPortAvailable(repr(self.RANDOM_PORT_RANGE))

        self.server = sock

    def accept(self):
        """
        accept client
        """
        try:
            (sock, _) = self.server.accept()
        except socket.timeout:
            return
        if self.transport_timeout:
            sock.settimeout(self.transport_timeout)
        self.sock = self.client = sock

    def get_port(self):
        return random.randint(*self.RANDOM_PORT_RANGE)

    def connected(self):
        return self.sock is not None

    def close(self):
        try:
            self.sock and self.sock.close()
            self.server and self.server.close()
        except:
            traceback.print_exc()


class NoPortAvailable(Exception):
    pass


class NoConnectedClient(Exception):
    pass



# -*- coding: utf-8 -*-

import os
from pylane.core.injector import inject as _inject


def inject(module, entrance, host, port, inject_args):
    """
    inject process with module.entrance
    """
    encoding = inject_args.get('encoding', 'utf-8')
    code = create_payload(module, entrance, host, port, encoding)
    inject_args['code'] = code
    success = _inject(**inject_args)
    if not success:
        exit(1)


def create_payload(module, entrance, host, port, encoding):
    """
    generate inject payload code for an exist file
    """
    base_path = os.path.dirname(os.path.abspath(__file__))
    file_path = base_path + '/%s.py' % module

    payload = "def start_remote_shell():\n"
    with open(file_path) as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            payload += '    ' + line
    payload += "\n    {entrance}(host='{host}', port={port}, encoding='{encoding}').start()\n".format(
        entrance=entrance,
        host=host,
        port=port,
        encoding=encoding
    )
    payload += '\nstart_remote_shell()'
    return payload



# -*- coding: utf-8 -*-

import sys
import traceback
import threading
import code
import json

from pylane.shell.sock import SockClient

if sys.version_info[0] == 3:
    from io import StringIO
else:
    from io import BytesIO as StringIO


class OutputHookContext(object):
    """
    Hook sys std with string io
    """

    def __enter__(self):
        self.sys_stdout = sys.stdout
        sys.stdout = self.stdout = StringIO()

        self.sys_stderr = sys.stderr
        sys.stderr = self.stderr = StringIO()

    def __exit__(self, *args, **kwargs):
        sys.stdout = self.sys_stdout
        sys.stderr = self.sys_stderr

    def truncate(self):
        for io in (sys.stdout, sys.stderr):
            io.seek(0)
            io.truncate()

    def getvalue(self):
        return (self.stdout.getvalue(), self.stderr.getvalue())

    def getvalue_truncate(self):
        value = (self.stdout.getvalue(), self.stderr.getvalue())
        self.truncate()
        return value


class RemoteShellThread(threading.Thread):
    """
    Start a thread in target process, run an interpreter
    """

    host = 'localhost'
    port = 9594
    debug = False

    def __init__(self, host=None, port=None, encoding='utf-8'):
        """
        """
        if host:
            self.host = host
        if port:
            self.port = port

        self.encoding = encoding
        self.lock = threading.RLock()
        self.stdio_hook = OutputHookContext()

        import __main__ as main
        self.user_defined_executor = getattr(
            main, '_user_defined_executor', None)
        self.in_user_defined_executor = False
        self.user_defined_executor_local = {}

        self._interpreter = code.InteractiveInterpreter(locals=locals())
        self.runsource('import sys')
        self.runsource('import __main__ as main')

        self.sock = SockClient(self.host, self.port, encoding)
        super(RemoteShellThread, self).__init__()

    def handle(self):
        """
        handle io
        """
        while True:
            cmd = self.sock.recv()
            # recv of sock closed returns '' in py2 and None in py3
            if cmd == '' or cmd is None:
                break
            flag = '0'
            self.sock.send(flag + self.runsource(cmd))

    def runsource(self, source, filename="<input>", symbol="single"):
        """
        run code in interpreter
        """
        with self.lock:
            source_strip = source.strip()
            if source_strip == '$enter_executor':
                if self.user_defined_executor:
                    self.in_user_defined_executor = True
                    return 'True'
                else:
                    return 'False, _user_defined_executor not found'
            if source_strip == '$exit_executor':
                self.in_user_defined_executor = False
                return 'True'

            if self.in_user_defined_executor and self.user_defined_executor:
                with self.stdio_hook:
                    return repr(self.user_defined_executor(
                        source_strip,
                        self.user_defined_executor_local
                    ))

            extend_function = self.get_extend_function(source_strip)
            if extend_function:
                with self.stdio_hook:
                    try:
                        return extend_function(self, source_strip)
                    except Exception as e:
                        print("Failed to run source", source, "err:", e)

            with self.stdio_hook:
                state, op = self._runsource(source)
                if not state:
                    print("Failed to run source", source)
                return op

    def get_output(self):
        """
        """
        out, err = self.stdio_hook.getvalue_truncate()
        try:
            out = out.decode(self.encoding)
            err = err.decode(self.encoding)
        except:
            # TODO
            pass
        return out, err

    def _runsource(self, source):
        # add \n in case of incomplete section code.
        ret = self._interpreter.runsource(
            source + "\n", filename="<input>", symbol="single")
        out, err = self.get_output()
        if ret is True:
            err += "error: cmd not completed."
        if err:
            ret = False, err
        else:
            ret = True, out
        if self.debug:
            open('/tmp/pmx-shell-log', 'w').write("%s\n%s" % (
                source, ret))
        return ret

    def run_help_doc(self, source):
        """
        support object?
        TODO more info when object??
        """
        item = source.replace('??', '').replace('?', '')
        state, doc = self._runsource('%s.__doc__' % item)
        if not state:
            return doc
        state, type_ = self._runsource('type(%s).__name__' % item)
        state, type_ = self._runsource('type(%s).__name__' % item)
        if not state:
            return type_
        op = "Type: %sDocstring: %s" % (type_, doc)
        return op

    def complete(self, source):
        """
        """
        text = source.split('$', 2)[1]
        suggs = []
        # for obj.\t
        if text.endswith('.'):
            item = text[:-1]
            state, op = self._runsource('dir(%s)' % item)
            if not state:
                return op
            dir_list = eval(op)
            suggs = [item + '.' + t for t in dir_list]
        # for obj.prefix\t
        elif '.' in text:
            item, sub_text = text.rsplit('.', 1)
            state, op = self._runsource('dir(%s)' % item)
            if not state:
                return op
            dir_list = eval(op)
            for t in dir_list:
                if t.startswith(sub_text):
                    suggs.append(item + '.' + t)
        # for prefix\t
        else:
            state, op = self._runsource('dir()')
            if not state:
                return op
            dir_list = eval(op)
            for t in dir_list:
                if t.startswith(text):
                    suggs.append(t)

        return json.dumps(suggs)

    def user_defined_function(self, source):
        import __main__ as main
        handler = getattr(main, '_user_defined_handler', None)\
            or getattr(main, '_telnet_handler', None)
        if not handler:
            return 'User defined handler not found'
        source = "main.%s('%s')" % (handler.__name__, source.replace('\n', ''))
        state, op = self._runsource(source)
        if not isinstance(op, str):
            op = repr(op)
        return op

    def get_extend_function(self, source):
        """
        """
        if source.endswith('?'):
            return self.extend_functions['help_doc']
        if source.startswith('$'):
            return self.extend_functions['udf']
        return self.extend_functions.get(
            source.split('$')[0]
        )

    extend_functions = {
        'help_doc': run_help_doc,
        'udf': user_defined_function,
        'complete': complete
    }

    def run(self):
        """
        main run entrance
        """
        self.name = "pylane-shell-thread"
        try:
            self.sock.connect()
            self.handle()
        except SystemExit:
            pass
        except:
            traceback.print_exc(file=sys.__stderr__)
        finally:
            if self.sock:
                self.sock.close()



# -*- coding: utf-8 -*-

import os
import sys
import signal
import time
import tempfile
import platform
import subprocess
import stat
import atexit
from threading import Timer
from .exception import (
    RequirementsInvalid,
    PylaneExceptionHandler
)


class Injector(object):
    """Inject a python process, run some code inside the vm."""

    def __init__(self,
                 pid=None,
                 code=None,
                 file_path=None,
                 gdb_path='gdb',
                 timeout=10,
                 verbose=0,
                 **ignore):
        """Init injector by args.
        Args:
            pid (int): target python process's pid.
            code (str): raw code.
            file_path (str): file path of the python code file.
            gdb_path (str): executable gdb path.
            timeout (int): timeout seconds.
            verbose (int): verbose level.

        Returns:
        """

        self.gdb = gdb_path
        self.timeout = timeout
        self.verbose = verbose

        self.env_detect()
        self.ensure_pid(pid)
        self.temp_file = None
        self.code_file = None
        self.ensure_code_file(code, file_path)

        atexit.register(self.cleanup)

    def env_detect(self):
        """"""
        # check platform
        self.env = env = {}
        if 'BSD' in platform.platform():
            env['bsd'] = True
            self.run = self._bsd_run
        if platform.dist()[0] == 'Ubuntu':
            env['ubuntu'] = True

        # check gdb
        if not os.access(self.gdb, os.X_OK):
            raise RequirementsInvalid(
                'gdb is not exist or not executable.'
            )
        # check ptrace
        ptrace_scope = '/proc/sys/kernel/yama/ptrace_scope'
        ptrace_req_msg = ('ptrace is disabled, enable it by:'
                          'echo 0 | sudo tee /proc/sys/kernel/yama/ptrace_scope . '
                          'arg --privileged may be also needed for docker '
                          'exec/run command to override ptrace_scope.')
        if os.path.exists(ptrace_scope):
            with open(ptrace_scope, 'r') as f:
                value = int(f.read().strip())
            if value == 1:
                raise RequirementsInvalid(ptrace_req_msg)
        else:
            getsebool = '/usr/sbin/getsebool'
            if os.path.exists(getsebool):
                p = subprocess.Popen(
                    [getsebool, 'deny_ptrace'],
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                out, err = p.communicate()
                if 'deny_ptrace --> on' in out:
                    raise RequirementsInvalid(ptrace_req_msg)

    def cleanup(self):
        """"""
        if self.temp_file and os.path.exists(self.temp_file):
            try:
                os.unlink(self.temp_file)
            except:
                pass

    def ensure_code_file(self, code, file_path):
        """"""
        if file_path:
            file_path = os.path.abspath(file_path)
            if not os.path.isfile(file_path):
                raise RequirementsInvalid(
                    'Arg file_path is not a valid file.'
                )
            self.code_file = file_path
        else:
            if code:
                (fd, temp_file_path) = tempfile.mkstemp()
                with os.fdopen(fd, 'w') as f:
                    f.write(code)
                self.code_file = self.temp_file = temp_file_path
            else:
                raise RequirementsInvalid(
                    'Neither code nor code file_path specified.'
                )
        st = os.stat(self.code_file)
        os.chmod(self.code_file,
                 st.st_mode | stat.S_IREAD | stat.S_IRGRP | stat.S_IROTH)

    def ensure_pid(self, pid):
        """"""
        ps_cmd = "ps a | grep -v grep | grep -w %s" % pid
        if not pid or not subprocess.getoutput(ps_cmd):
            raise RequirementsInvalid('Process %s not exist.' % pid)
        self.pid = pid

    def generate_gdb_codes(self):
        """Generate gdb command codes
        Args:
        Returns:
            list: gdb command code lines
        """

        lib_path = os.path.abspath(
            os.path.join(os.path.abspath(__file__), '../..')
        )
        inject_paths = [
            # TODO to support injected code file's path
            # cannot support docker container
            # os.path.dirname(self.code_file),
            # current path, not used
            os.getcwd(),
            lib_path
        ]
        temp_sys_name = '_pylane_sys'
        prepare_code = ' '.join([
            'import sys as %s;' % temp_sys_name,
            '%s.path.extend([%s]);' % (
                temp_sys_name,
                ','.join(['\\\"%s\\\"' % p for p in inject_paths])
            )
        ])
        # deprecated, cause its hard to pass python code in shell args
        # exec code is passed by shell command line, so add \\\" and \\\\n
        # run_code = 'exec(\\\"%s\\\")' % (self.code.replace('\n', '\\\\n'))
        if sys.version_info.major == 2:
            run_code = ' '.join([
                '__code_file = open(\\\"%s\\\");' % self.code_file,
                '__raw_code = __code_file.read();',
                '__code_file.close();',
                # python 2 donot support exec as Thread's target param
                'exec(__raw_code);',
                'del __code_file;',
                'del __raw_code;'
            ])
        else:
            run_code = ' '.join([
                '__code_file = open(\\\"%s\\\");' % self.code_file,
                '__raw_code = __code_file.read();',
                '__code_file.close();',
                # run code async and stop injection early to keep target process safe
                'from threading import Thread as __Thread;'
                '__thread = __Thread(target=exec, args=(__raw_code,));'
                '__thread.daemon = True;'
                '__thread.start();'
                'del __code_file;'
                'del __raw_code;'
                'del __Thread;'
                'del __thread;'
            ])
        # TODO injected code may change path as well
        cleanup_code = ' '.join([
            '%s.path = %s.path[:-%s];' % (
                temp_sys_name,
                temp_sys_name,
                len(inject_paths)
            )
        ])
        return [
            # use char in case of symbol PyGilState_STATE not found
            'call $gil_state = (char) PyGILState_Ensure()',
            'call (void) PyRun_SimpleString("%s")' % prepare_code,
            'call (void) PyRun_SimpleString("%s")' % run_code,
            'call (void) PyRun_SimpleString("%s")' % cleanup_code,
            # make sure previous codes are safe.
            # gdb exit without GIL release is a disaster for target process.
            'call (void) PyGILState_Release($gil_state)',
        ]

    def timeout_exit(self, process):
        print("timeout in %s secs, exit." % self.timeout)
        os.kill(process.pid, signal.SIGTERM)

    def inject(self):
        """Run inject"""
        codes = self.generate_gdb_codes()
        process = self.run(codes)
        timer = Timer(self.timeout, self.timeout_exit, (process,))
        out = b''
        err = b''
        try:
            timer.start()
            out, err = process.communicate()
        finally:
            timer.cancel()
            self.cleanup()
            if self.verbose:
                print('stdout:', out)
                print('stderr:', err)
                print(err)
            if b'Operation not permitted' in err:
                print('Cannot attach a process without perm.')
                if self.env.get('ubuntu'):
                    print('You may need root perm to use ptrace in Ubuntu.')
                exit(1)
        return True

    def _bsd_run(self, codes):
        """gdb under bsd can only run command in a file"""

        (fd, gdb_file_name) = tempfile.mkstemp()
        with os.fdopen(fd, 'w') as f:
            f.write('\n'.join(codes) + '\nquit')

        command = "%s --command=%s --pid=%s --exec=python" % (
            self.gdb, gdb_file_name, self.pid)

        p = self._popen(command)
        # sleep 0.5 seconds to avoid removing gdb_file too fast
        time.sleep(0.5)
        os.remove(gdb_file_name)
        return p

    def _run(self, codes):
        """gdb under linux"""

        batch_commands = ' '.join(
            ["-eval-command='%s'" % code for code in codes]
        )
        command = '%s -p %d -batch %s' % (self.gdb, self.pid, batch_commands)

        return self._popen(command)

    run = _run

    def _popen(self, command):
        """shell popen with pipe std"""
        return subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )


@PylaneExceptionHandler
def inject(*args, **kwargs):
    return Injector(*args, **kwargs).inject()



# -*- coding: utf-8 -*-

"""
"""

from .proxy import IPythonShellProxy as ShellProxy


class Shell(object):
    """Shell server"""

    # pre_codes defined to run before shell is available
    pre_codes = [
        """
            print('Command: %s %s %s' % (
                sys.executable,
                getattr(main, '__file__', '__main__'),
                ' '.join([
                    '"%s"' % (
                        arg.replace('\\\\', '\\\\\\\\').replace('"', '\\\\"'),
                    )
                    for arg in sys.argv[1:]
                ]),
            ))
        """,
        "import pylane.shell.help_functions as tools",
        # "from imp import reload",
        # "tools = reload(tools)",
    ]

    # extend help infos followed by ipython help info
    extend_help = """Welcome to pylane shell. Features below are avaliable:
    Use TAB to get remote auto completion.
    Use 'main' to access remote process's __main__ object.
    Use 'tools' to get pylane toolkit functions.
        Example: tools.get_insts(YOUR_CLASS_NAME).
    Use 'a_object?' for more usage info.
    """
    udf_help = """
    User Defined Executor:
        Put a _user_defined_executor(input, local) function in __main__ global,
            then use $enter_executor to enter, $exit_executor to exit.
        When in executor, all input will be handled by executor.
        Arg local is an empty dict.
    User Defined Functions:
        Put a _user_defined_handler($cmd) function in __main__ global,
            then use $cmd to call them.
    """

    def __init__(self,
                 **inject_args):
        """Init injector by args.
        Args:
            encoding (str): encoding used when shell transporting

        Returns:
        """
        self.inject_args = inject_args
        self.encoding = inject_args.get('encoding')

    def run(self):
        """run shell server and connect"""
        shell_proxy = ShellProxy(self.inject_args)
        shell_proxy.run()
        shell_proxy.runsource("print('''%s''')" % self.extend_help)
        for pre_code in self.pre_codes:
            code = self.raw_code(pre_code)
            shell_proxy.runsource(code)
        shell_proxy.interact()

    def raw_code(self, code):
        """parse code in docstring"""
        lines = code.split('\n')
        for start_line in lines:
            if start_line:
                break
        blanks = len(start_line) - len(start_line.lstrip())
        code = '\n'.join([
            line[blanks:]
            for line in lines
        ])
        if not code.endswith('\n'):
            code += '\n'
        return code


def shell(*args, **kwargs):
    return Shell(*args, **kwargs).run()



# -*- coding: utf-8 -*-


import time
from threading import Thread
from .sock import SockServer
from .inject import inject
from .ipython_embed import IPythonShell


class AsyncServer(object):
    """
    """
    LISTEN_TIMEOUT = 30

    def start_server(self, encoding, timeout):
        """start sock server and wait for client"""
        self.sock_server = SockServer(
            listen_timeout=self.LISTEN_TIMEOUT,
            transport_timeout=timeout or 0,
            encoding=encoding)
        self.sock_server.listen()
        self.async_accept()
        return (self.sock_server.host, self.sock_server.port)

    def async_accept(self):
        """start a thread to wait for sock server accept"""
        thread = Thread(
            target=self.sock_server.accept, name="sock_server_async_accept")
        thread.daemon = True
        self.async_accept_thread = thread
        thread.start()

    def wait(self):
        """wait for remote client to connet"""
        while self.async_accept_thread and self.async_accept_thread.is_alive():
            time.sleep(0.1)
        if not self.sock_server.connected():
            raise NoConnectedClient(
                "No client connected in %s seconds" % self.LISTEN_TIMEOUT
            )
        self.sock = self.sock_server


class NoConnectedClient(Exception):
    pass


class SimpleRemoteProxy(object):
    MODULE = 'remote_shell'
    ENTRANCE = 'RemoteShellThread'


class BaseProxy(AsyncServer, SimpleRemoteProxy):
    """"""

    def __init__(self, inject_args):
        """"""
        self.inject_args = inject_args

    def run(self):
        """"""
        host, port = self.start_server(
            encoding=self.inject_args.get('encoding'),
            timeout=self.inject_args.get('timeout')
        )
        inject(
            module=self.MODULE,
            entrance=self.ENTRANCE,
            host=host,
            port=port,
            inject_args=self.inject_args)
        self.wait()


class IPythonShellProxy(BaseProxy, IPythonShell):
    """"""

    def __init__(self, inject_args):
        BaseProxy.__init__(self, inject_args)
        IPythonShell.__init__(self, inject_args.get('encode'))



# -*- coding: utf-8 -*-

import json
from IPython.terminal.embed import InteractiveShellEmbed
from IPython.core.interactiveshell import ExecutionResult


class IPythonShell(InteractiveShellEmbed):
    """use an ipython interface to provide shell"""

    ENCODING = 'utf-8'

    def __init__(self, encoding='utf-8'):
        InteractiveShellEmbed.__init__(self)
        self.set_custom_completer(self.remote_completer, pos=0)
        self.encoding = encoding or self.ENCODING

    def ipython_pre_works(self, result, raw_cell, store_history,
                          silent=False, shell_futures=True,
                          ):
        if silent:
            store_history = False

        if store_history:
            result.execution_count = self.execution_count

        def error_before_exec(value):
            result.error_before_exec = value
            return result

        self.events.trigger('pre_execute')
        if not silent:
            self.events.trigger('pre_run_cell')

        preprocessing_exc_tuple = None
        cell = raw_cell

        if store_history:
            self.history_manager.store_inputs(self.execution_count,
                                              cell, raw_cell)
        if not silent:
            self.logger.log(cell, raw_cell)

        if preprocessing_exc_tuple is not None:
            self.showtraceback(preprocessing_exc_tuple)
            if store_history:
                self.execution_count += 1
            return error_before_exec(preprocessing_exc_tuple[2])

        return cell

    def ipython_post_works(self, result, store_history):
        if store_history:
            self.history_manager.store_output(self.execution_count)
            self.execution_count += 1
            result.execution_count = self.execution_count

    def redirect_to_ipython(self, raw_cell):
        """
        to be finished
        """
        if raw_cell == '?\n':
            return True
        if raw_cell.startswith('%'):
            return True

        return False

    def run_cell(self, raw_cell, store_history=True,
                 silent=False, shell_futures=True
                 ):
        """
        run ipython code in ipython run_cell,
        run python console code in remoute console
        """
        if self.redirect_to_ipython(raw_cell):
            return InteractiveShellEmbed.run_cell(
                self, raw_cell, store_history, silent, shell_futures
            )

        result = ExecutionResult()
        self.displayhook.exec_result = result

        if (not raw_cell) or raw_cell.isspace():
            return result

        cell = self.ipython_pre_works(
            result, raw_cell, store_history, silent, shell_futures
        )

        output = self._runsource(cell)
        if output:
            result.result = output
            self._output(output)
            # self.write(output)
        self.ipython_post_works(result, store_history)

        return result

    def _output(self, output):
        """
        try to display evaled object
        if failed, write raw string
        """
        try:
            output = eval(output)
            try:
                output = output.decode(self.encoding)
            except:
                pass
            self.displayhook(output)
            return
        except:
            pass
        if not output.endswith('\n\n'):
            output += '\n'
        self.displayhook.start_displayhook()
        self.displayhook.write_output_prompt()
        self.write('\n' + output)
        self.displayhook.finish_displayhook()

    def runsource(self, source):
        self.write(self._runsource(source))

    def _runsource(self, source, *args, **kwargs):
        self.sock.send(source)
        raw = self.sock.recv()
        if raw == '' or raw is None:
            exit(0)
        flag, op = raw[0], raw[1:]
        # TODO old interface which is not used, going to be deprecated.
        flag is '0'
        return op

    def remote_completer(self, ipcompleter, text, line=None, cursor_pos=None):
        source = 'complete$%s' % text
        op = self._runsource(source)
        try:
            # ignore those cases like obj not exists
            return json.loads(op)
        except:
            return []



#!/usr/bin/env python
# encoding: utf-8

'''
All functions here must be written with function doc.
And all exceptions must be handled in this module....
Don't Import Any Unnecessary things.
Use remote ipc to inject, don't import them in remote process in case of reload.
'''

from __future__ import print_function


def get_insts(class_name):
    """get_insts(class_name) -> instance_list"""
    ins_list = []
    import gc
    for obj in gc.get_objects():
        try:
            if getattr(obj, "__class__", None):
                if obj.__class__.__name__ == class_name:
                    ins_list.append(obj)
        # weak ref items will raise an exc here
        except ReferenceError:
            continue
    return ins_list


def get_classes(class_name):
    """get_classes(class_name) -> class_object_list : cast new type class only"""
    cls_list = []
    import gc
    import inspect
    for obj in gc.get_objects():
        try:
            if getattr(obj, "__class__", None):
                if inspect.isclass(obj) and obj.__name__ == class_name:
                    cls_list.append(obj)
        # weak ref items will raise an exc here
        except ReferenceError:
            continue
    return cls_list


def cast_id(object_id):
    """cast_id(object_id) -> object"""
    if not getattr(cast_id, 'ensured', False):
        print("""Warning: If you provide an invalid or expired id, this call will cause a segment fault,
the attached process will be killed.
cast_id will do nothing this time. If you know whats going on, recall this function to execute.""")
        setattr(cast_id, 'ensured', True)
        return
    import ctypes
    return ctypes.cast(object_id, ctypes.py_object).value


def get_source_code(obj):
    """get_source_code(obj) -> source_code_string : print it for pretty format"""
    import inspect
    try:
        return ''.join(inspect.getsourcelines(obj)[0])
    except Exception as e:
        print('failed, error:', e)


def print_source_code(obj):
    """print_source_code(obj) -> None : print source code."""
    print(get_source_code(obj))


def inspect_threads(thread_names=[]):
    """inspect_threads() -> {thread_name: {"locals": {}, "stack": ""}} : return threads' locals and stack"""
    import threading
    import sys
    import traceback
    pylane_thread_name = "pylane-shell-thread"
    stacks = {}
    frames = sys._current_frames()
    threads = threading.enumerate()
    for thread in threads:
        if thread.name == pylane_thread_name:
            continue
        if thread_names and thread.name not in thread_names:
            continue
        frame = frames.get(thread.ident)
        stack = ''.join(traceback.format_stack(frame)) if frame else ''
        stacks[thread.name] = {
            "locals": frame.f_locals,
            "stack": stack
        }
    return stacks


def print_threads(thread_names=[]):
    """print_threads() -> None : print threads' stack and locals"""
    stacks = inspect_threads(thread_names)
    for name, thread_info in stacks.items():
        print("\nThread:", name)
        print("Stack:\n", thread_info['stack'])
        print("Locals:\n", thread_info['locals'])


def log_it(func):
    """log_it(func) -> decoratored_func : a decorator to print func input and output in log"""
    import logging
    import time

    def wrapper(*args, **kwargs):
        start = time.time()
        ret = func(*args, **kwargs)
        duration = time.time() - start
        logging.warning("func: %s sec_cost: %s input: %s output %s" % (
            func.__name__, round(duration, 4), repr((args, kwargs)), repr(ret)
        ))
        return ret
    return wrapper



# -*- coding: utf-8 -*-


import traceback


class PylaneException(Exception):
    pass


class RequirementsInvalid(PylaneException):
    pass


def PylaneExceptionHandler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except PylaneException as e:
            for arg in e.args:
                print(arg)
        # TODO
        except Exception as e:
            print('Internal error occured, contact author if u need help.')
            for arg in e.args:
                print(arg)
            print(traceback.format_exc())
        exit(1)
    return wrapper



# -*- coding: utf-8 -*-


from setuptools import setup
from os import path


VERSION = '0.0.7'

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
    long_description = f.read()


setup(
    name='pylane',
    version=VERSION,
    author='valensc, Wu Xiao',
    author_email='weidong1312@gmail.com, notgiven@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/NtesEyes/pylane',
    download_url='https://github.com/NtesEyes/pylane/archive/0.0.2.tar.gz',
    packages=[
        'pylane',
        'pylane.core',
        'pylane.shell',
    ],
    entry_points="""
          [console_scripts]
          pylane = pylane.entry:main
      """,
    # cmdclass={
        # 'build_py': build_py
    # },
    install_requires=[
        "ipython==5.7",
        # 'ipython==5.8;python_version<"3.4"',
        # 'ipython==7.2;python_version>="3.4"',
        "Click==7.0",
    ],
    keywords=['debug', 'attach', 'gdb', 'shell']
)



# -*- coding: utf-8 -*-

"""
Entry of Pylane
"""

import click
from pylane.core.injector import inject as _inject
from pylane.shell.shell import shell as _shell


@click.command()
@click.argument('pid', type=int)
@click.argument('file_path', type=click.Path(exists=True, readable=True))
@click.pass_context
def inject(ctx, pid, file_path):
    """Inject a python process, run some code inside the vm."""
    _inject(pid=pid, file_path=file_path, **ctx.obj)


@click.command()
@click.argument('pid', type=int)
@click.pass_context
def shell(ctx, pid):
    """Create a remote python shell on a process."""
    _shell(pid=pid, **ctx.obj)


@click.group()
@click.option('-g', '--gdb_path', type=click.Path(exists=True), default='/usr/bin/gdb')
@click.option('-v', '--verbose', count=True)
@click.option('-t', '--timeout', type=int, default=10)
@click.option('-e', '--encoding', type=str, default='utf-8')
@click.pass_context
def main_entry(ctx, **kwargs):
    ctx.obj = kwargs


main_entry.add_command(inject)
main_entry.add_command(shell)


def main():
    main_entry()


if __name__ == '__main__':
    main()
